/*	Copyright (C) 2003-2013 Free Electron Organization
	Any use of this software requires a license.  If a valid license
	was not distributed with this file, visit freeelectron.org. */

/** @file */

#include "surface.pmh"

#define	FE_SAC_MAX_VERTS	4	//* 0 = any number
#define	FE_SAC_CHECK_LIMIT	FALSE

namespace fe
{

SurfaceAccessorCatalog::SurfaceAccessorCatalog(void):
	m_pStringList(NULL),
	m_pIntegerList(NULL),
	m_pRealList(NULL),
	m_pVectorList(NULL)
{
	setName("SurfaceAccessorCatalog");
}

SurfaceAccessorCatalog::~SurfaceAccessorCatalog(void)
{
}

void SurfaceAccessorCatalog::setAttrType(String a_attrType)
{
	if(m_attrType==a_attrType)
	{
		return;
	}
	if(!m_attrType.empty())
	{
		feLog("SurfaceAccessorCatalog::setAttrType"
				" changing type from \"%s\" to \"%s\"\n",
				m_attrType.c_str(),a_attrType.c_str());
	}

	m_attrType=a_attrType;

	m_spCatalog->catalog<String>(m_key,"type:"+m_attrName)=a_attrType;

	if(m_attrType=="string")
	{
		m_pStringList=&m_spCatalog->catalog< std::vector<String> >(
				m_key,m_attrName);
	}
	else if(m_attrType=="integer")
	{
		m_pIntegerList=&m_spCatalog->catalog< std::vector<I32> >(
				m_key,m_attrName);
	}
	else if(m_attrType=="real")
	{
		m_pRealList=&m_spCatalog->catalog< std::vector<Real> >(
				m_key,m_attrName);
	}
	else if(m_attrType=="vector3")
	{
		String rename=m_attrName;

		if(m_attribute==SurfaceAccessibleI::e_vertices)
		{
			rename="spc:at";

			m_pIntegerList=&m_spCatalog->catalog< std::vector<I32> >(
					m_key,m_attrName);
		}

		m_pVectorList=&m_spCatalog->catalog< std::vector<SpatialVector> >(
				m_key,rename);
	}
}

U32 SurfaceAccessorCatalog::count(void) const
{
	switch(m_element)
	{
		case SurfaceAccessibleI::e_point:
			return m_spCatalog->catalogOrDefault<I32>(m_key,":point_count",0);
			break;
		case SurfaceAccessibleI::e_vertex:
			break;
		case SurfaceAccessibleI::e_primitive:
			return m_spCatalog->catalogOrDefault<I32>(m_key,":prim_count",0);
			break;
		case SurfaceAccessibleI::e_detail:
			break;
		default:
			;
	}

	return 0;
}

U32 SurfaceAccessorCatalog::subCount(U32 a_index) const
{
	if(m_attribute==SurfaceAccessibleI::e_vertices)
	{
#if	FE_SAC_MAX_VERTS
		const_cast<SurfaceAccessorCatalog*>(this)->setAttrType("vector3");

		FEASSERT(m_pIntegerList);

		const I32 grain=(FE_SAC_MAX_VERTS+1);
		const I32 offset=a_index*grain;

		const U32 size=m_pIntegerList->size();
		if(offset>=size)
		{
			return 0;
		}

		//* first index in granule is number of entries
		return (*m_pIntegerList)[offset];
#else
		const std::vector< std::vector<I32> >& rIntegerIntegerList=
				m_spCatalog->catalog<std::vector< std::vector<I32> > >(
				m_key,m_attrName);

		const U32 size=rIntegerIntegerList.size();
		if(a_index>=size)
		{
			return 0;
		}

		const std::vector<I32>& rIntegerList=
				rIntegerIntegerList[a_index];
		return rIntegerList.size();
#endif
	}

	return 0;
}

void SurfaceAccessorCatalog::set(U32 a_index,U32 subIndex,String a_string)
{
	setAttrType("string");

#if FE_SAC_CHECK_LIMIT
	if(a_index>=count())
	{
		return;
	}
#endif

	FEASSERT(m_pStringList);

	const U32 size=m_pStringList->size();
	if(a_index>=size)
	{
		m_pStringList->resize(fe::maximum(a_index+1,count()));
	}
	(*m_pStringList)[a_index]=a_string;

	for(U32 index=size;index<a_index;index++)
	{
		(*m_pStringList)[index]="";
	}
}

String SurfaceAccessorCatalog::string(U32 a_index,U32 a_subIndex)
{
	setAttrType("string");

	FEASSERT(m_pStringList);

	if(a_index<m_pStringList->size())
	{
		return (*m_pStringList)[a_index];
	}

	return "";
}

void SurfaceAccessorCatalog::set(U32 a_index,U32 a_subIndex,I32 a_integer)
{
#if FE_SAC_CHECK_LIMIT
	if(a_index>=count())
	{
		return;
	}
#endif

	if(m_attribute==SurfaceAccessibleI::e_vertices)
	{
#if	FE_SAC_MAX_VERTS
		if(a_subIndex>=FE_SAC_MAX_VERTS)
		{
			feX("SurfaceAccessorCatalog::set","subIndex %d exceeds max %d",
					a_subIndex,FE_SAC_MAX_VERTS-1);
			return;
		}

		setAttrType("vector3");

		FEASSERT(m_pIntegerList);

		const I32 grain=(FE_SAC_MAX_VERTS+1);
		const I32 offset=a_index*grain;

		const U32 size=m_pIntegerList->size();
		if(offset>=size)
		{
			m_pIntegerList->resize(
					fe::maximum(offset+grain,I32(count()*grain)));
		}

		//* first index in granule is number of entries
		I32& rEntryCount=(*m_pIntegerList)[offset];
		if(rEntryCount<a_subIndex+1)
		{
			rEntryCount=a_subIndex+1;
		}
		(*m_pIntegerList)[offset+1+a_subIndex]=a_integer;
#else
		std::vector< std::vector<I32> >& rIntegerIntegerList=
				m_spCatalog->catalog< std::vector< std::vector<I32> > >(
				m_key,m_attrName);

		const U32 size=rIntegerIntegerList.size();
		if(a_index>=size)
		{
			rIntegerIntegerList.resize(fe::maximum(a_index+1,count()));
		}

		std::vector<I32>& rIntegerList=rIntegerIntegerList[a_index];

		const U32 subSize=rIntegerList.size();
		if(a_subIndex>=subSize)
		{
			rIntegerList.resize(a_subIndex+1);
		}
		rIntegerList[a_subIndex]=a_integer;
#endif
		return;
	}

	setAttrType("integer");

	FEASSERT(m_pIntegerList);

	const U32 size=m_pIntegerList->size();
	if(a_index>=size)
	{
		m_pIntegerList->resize(fe::maximum(a_index+1,count()));
	}
	(*m_pIntegerList)[a_index]=a_integer;

	for(U32 index=size;index<a_index;index++)
	{
		(*m_pIntegerList)[index]=0;
	}
}

I32 SurfaceAccessorCatalog::integer(U32 a_index,U32 a_subIndex)
{
	if(m_attribute==SurfaceAccessibleI::e_vertices)
	{
#if	FE_SAC_MAX_VERTS
		if(a_subIndex>=FE_SAC_MAX_VERTS)
		{
			feX("SurfaceAccessorCatalog::integer","subIndex %d exceeds max %d",
					a_subIndex,FE_SAC_MAX_VERTS-1);
			return -1;
		}

		setAttrType("vector3");

		FEASSERT(m_pIntegerList);

		const I32 grain=(FE_SAC_MAX_VERTS+1);
		const I32 offset=a_index*grain;

		const U32 size=m_pIntegerList->size();
		if(offset>=size)
		{
			return -1;
		}

		//* first index in granule is number of entries
		const U32 entryCount=(*m_pIntegerList)[offset];
		if(a_subIndex>=entryCount)
		{
			return -1;
		}
		return (*m_pIntegerList)[offset+1+a_subIndex];
#else
		const std::vector< std::vector<I32> >& rIntegerIntegerList=
				m_spCatalog->catalog< std::vector< std::vector<I32> > >(
				m_key,m_attrName);

		const std::vector<I32>& rIntegerList=rIntegerIntegerList[a_index];
		return rIntegerList[a_subIndex];
#endif
	}

	setAttrType("integer");

	FEASSERT(m_pIntegerList);

	if(a_index<m_pIntegerList->size())
	{
		return (*m_pIntegerList)[a_index];
	}

	return 0;
}

//* coordinate with SurfaceAccessorBase::append(I32)
I32 SurfaceAccessorCatalog::append(I32 a_integer)
{
	//* lot of points or primitives
	if(m_element==SurfaceAccessibleI::e_point ||
			(m_element==SurfaceAccessibleI::e_primitive &&
			m_attribute!=SurfaceAccessibleI::e_vertices))
	{
		if(a_integer<1)
		{
			return -1;
		}

		const I32 first=count();

		switch(m_element)
		{
			case SurfaceAccessibleI::e_point:
				m_spCatalog->catalog<I32>(m_key,":point_count")=first+a_integer;
				break;
			case SurfaceAccessibleI::e_vertex:
				break;
			case SurfaceAccessibleI::e_primitive:
				m_spCatalog->catalog<I32>(m_key,":prim_count")=first+a_integer;
				break;
			case SurfaceAccessibleI::e_detail:
				break;
			default:
				return -1;
		}

		return first;
	}

	if(m_element!=SurfaceAccessibleI::e_primitive ||
			m_attribute!=SurfaceAccessibleI::e_vertices)
	{
		append(U32(0),a_integer);
		return 0;
	}

	//* append a primitive with the specified number of vertices
	sp<SurfaceAccessorI> spPointAccessor=m_spSurfaceAccessibleI->accessor(
			SurfaceAccessibleI::e_point,SurfaceAccessibleI::e_position);

	const I32 primitiveIndex=append();
	for(I32 vertIndex=0;vertIndex<a_integer;vertIndex++)
	{
		const I32 pointIndex=spPointAccessor->append();
		append(primitiveIndex,pointIndex);
	}

	return primitiveIndex;
}

I32 SurfaceAccessorCatalog::append(void)
{
	//* point or primitive

	switch(m_element)
	{
		case SurfaceAccessibleI::e_point:
		{
			return m_spCatalog->catalog<I32>(m_key,":point_count")++;
		}
		case SurfaceAccessibleI::e_vertex:
			break;
		case SurfaceAccessibleI::e_primitive:
		{
			return m_spCatalog->catalog<I32>(m_key,":prim_count")++;
		}
		case SurfaceAccessibleI::e_detail:
			break;
		default:
			return -1;
	}

	return count();
}

void SurfaceAccessorCatalog::append(U32 a_index,I32 a_integer)
{
	//* vertex of primitive
	const I32 vertIndex=subCount(a_index);
	set(a_index,vertIndex,a_integer);
}

//* coordinate with SurfaceAccessorBase::append(a_rPrimVerts)
void SurfaceAccessorCatalog::append(
	std::vector< std::vector<I32> >& a_rPrimVerts)
{
	if(m_element!=SurfaceAccessibleI::e_primitive ||
			m_attribute!=SurfaceAccessibleI::e_vertices)
	{
		return;
	}

	const U32 appendCount=a_rPrimVerts.size();

	I32 primitiveIndex=
			m_spCatalog->catalogOrDefault<I32>(m_key,":prim_count",0);
	m_spCatalog->catalog<I32>(m_key,":prim_count")=primitiveIndex+appendCount;

#if	FE_SAC_MAX_VERTS
	const I32 grain=(FE_SAC_MAX_VERTS+1);

	setAttrType("vector3");

	FEASSERT(m_pIntegerList);

	for(U32 appendIndex=0;appendIndex<appendCount;appendIndex++)
	{
		const std::vector<I32>& rVertList=a_rPrimVerts[appendIndex];

		const I32 offset=primitiveIndex*grain;

		const U32 size=m_pIntegerList->size();
		if(offset>=size)
		{
			m_pIntegerList->resize(
					fe::maximum(offset+grain,I32(count()*grain)));
		}

		const U32 subCount=fe::minimum(I32(rVertList.size()),FE_SAC_MAX_VERTS);

		//* first index in granule is number of entries
		(*m_pIntegerList)[offset]=subCount;
		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			(*m_pIntegerList)[offset+1+subIndex]=rVertList[subIndex];
		}
		primitiveIndex++;
	}
#else
	for(U32 appendIndex=0;appendIndex<appendCount;appendIndex++)
	{
		const std::vector<I32>& rVertList=a_rPrimVerts[appendIndex];

		const U32 subCount=rVertList.size();
		for(U32 subIndex=0;subIndex<subCount;subIndex++)
		{
			const I32 pointIndex=rVertList[subIndex];

			//* TODO direct
			append(primitiveIndex,pointIndex);	//* add vertex to primitive
		}
		primitiveIndex++;
	}
#endif

}

void SurfaceAccessorCatalog::set(U32 a_index,U32 a_subIndex,Real a_real)
{
	setAttrType("real");

	FEASSERT(m_pRealList);

#if FE_SAC_CHECK_LIMIT
	if(a_index>=count())
	{
		return;
	}
#endif

	const U32 size=m_pRealList->size();
	if(a_index>=size)
	{
		m_pRealList->resize(fe::maximum(a_index+1,count()));
	}
	(*m_pRealList)[a_index]=a_real;

	for(U32 index=size;index<a_index;index++)
	{
		(*m_pRealList)[index]=0.0;
	}
}

Real SurfaceAccessorCatalog::real(U32 a_index,U32 a_subIndex)
{
	setAttrType("real");

	FEASSERT(m_pRealList);

	if(a_index<m_pRealList->size())
	{
		return (*m_pRealList)[a_index];
	}

	return 0.0;
}

void SurfaceAccessorCatalog::set(U32 a_index,U32 a_subIndex,
								const SpatialVector& a_spatialVector)
{
	setAttrType("vector3");

	FEASSERT(m_pVectorList);

#if FE_SAC_CHECK_LIMIT
	if(a_index>=count())
	{
		return;
	}
#endif

	const U32 size=m_pVectorList->size();
	if(a_index>=size)
	{
		m_pVectorList->resize(fe::maximum(a_index+1,count()));
	}
	(*m_pVectorList)[a_index]=a_spatialVector;

	for(U32 index=size;index<a_index;index++)
	{
		fe::set((*m_pVectorList)[index]);
	}

	return;
}
SpatialVector SurfaceAccessorCatalog::spatialVector(
	U32 a_index,U32 a_subIndex)
{
	setAttrType("vector3");

	FEASSERT(m_pVectorList);

	U32 reIndex=a_index;
	if(m_attribute==SurfaceAccessibleI::e_vertices)
	{
		const I32 pointIndex=integer(a_index,a_subIndex);

		reIndex=pointIndex;
	}

	if(reIndex<m_pVectorList->size())
	{
		return (*m_pVectorList)[reIndex];
	}

	return SpatialVector(0.0,0.0,0.0);
}

BWORD SurfaceAccessorCatalog::bindInternal(
	SurfaceAccessibleI::Element a_element,const String& a_name)
{
	m_element=a_element;
	m_attrName=a_name;

	if(!m_spCatalog.isValid())
	{
		feLog("SurfaceAccessorCatalog::bindInternal invalid Catalog\n");
		return FALSE;
	}

	return TRUE;
}

} // namespace
